#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-09-13 10:30:05
Last Modified by: Hanyu Wang
Last Modified time: 2023-09-13 10:31:53
"""

from scipy.linalg import det

import numpy as np


def to_special_unitary(matrix: np.ndarray) -> np.ndarray:
    """Convert gate tensor to the special unitary group."""
    rank = matrix.shape[0]
    matrix_ = matrix / det(matrix) ** (1 / rank)
    return matrix_


def unitary_zyz_decomposition(matrix: np.ndarray):
    """
    Returns the Euler Z-Y-Z decomposition of a local 1-qubit gate.
    """
    U = to_special_unitary(matrix)
    if abs(U[0, 0]) > abs(U[1, 0]):
        theta1 = 2 * np.arccos(min(abs(U[0, 0]), 1))
    else:
        theta1 = 2 * np.arcsin(min(abs(U[1, 0]), 1))
    cos_halftheta1 = np.cos(theta1 / 2)
    if not np.isclose(cos_halftheta1, 0.0):
        phase = U[1, 1] / cos_halftheta1
        theta0_plus_theta2 = 2 * np.arctan2(np.imag(phase), np.real(phase))
    else:
        theta0_plus_theta2 = 0.0
    sin_halftheta1 = np.sin(theta1 / 2)
    if not np.isclose(sin_halftheta1, 0.0):
        phase = U[1, 0] / sin_halftheta1
        theta0_sub_theta2 = 2 * np.arctan2(np.imag(phase), np.real(phase))
    else:
        theta0_sub_theta2 = 0.0
    theta0 = (theta0_plus_theta2 + theta0_sub_theta2) / 2
    theta2 = (theta0_plus_theta2 - theta0_sub_theta2) / 2
    # this is very important, otherwise the result will be wrong
    if np.isclose(theta1, 0.0, atol=1e-6):
        theta2 = theta0 + theta2
        theta1 = 0.0
        theta0 = 0.0
    if theta0 < 0:
        theta0 += 2 * np.pi
    if theta1 < 0:
        theta1 += 2 * np.pi
    if theta2 < 0:
        theta2 += 2 * np.pi

    return theta0, theta1, theta2


class UnitaryGate:
    """Classmethod to handle RotationGate ."""

    def __init__(self, unitary: np.ndarray) -> None:
        self.unitary = unitary

        self.alpha, self.beta, self.gamma = unitary_zyz_decomposition(unitary)

    def get_unitary(self) -> float:
        """Get the theta of the rotation gate .

        :return: [description]
        :rtype: float
        """
        return self.unitary

    def get_alpha(self) -> float:
        """Get the theta of the rotation gate .

        :return: [description]
        :rtype: float
        """
        return self.alpha

    def get_beta(self) -> float:
        """Get the theta of the rotation gate .

        :return: [description]
        :rtype: float
        """
        return self.beta

    def get_gamma(self) -> float:
        """Get the theta of the rotation gate .

        :return: [description]
        :rtype: float
        """
        return self.gamma

    def is_z_trivial(self) -> bool:
        """Whether the model is trivial to be trivial .

        :return: [description]
        :rtype: bool
        """
        is_alpha_trivial = np.isclose(self.alpha, 0) or np.isclose(
            self.alpha, 2 * np.pi
        )
        is_gamma_trivial = np.isclose(self.gamma, 0) or np.isclose(
            self.gamma, 2 * np.pi
        )

        return is_alpha_trivial and is_gamma_trivial

    def is_z_trivial_not(self) -> bool:
        """Whether z - T is z - T .

        :return: [description]
        :rtype: bool
        """
        is_alpha_trivial = np.isclose(self.alpha, np.pi)
        is_gamma_trivial = np.isclose(self.gamma, np.pi)

        return is_alpha_trivial and is_gamma_trivial



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-09-15 13:14:21
Last Modified by: Hanyu Wang
Last Modified time: 2023-09-15 14:32:11
"""

from collections import namedtuple
from typing import List
import numpy as np


from xyz.circuit import QCircuit, RY, QGate
from xyz.qstate import QState
from xyz.utils import stopwatch
from xyz.utils import global_stopwatch_report
from xyz.utils.colors import print_yellow

from ._exact_cnot_synthesis import exact_cnot_synthesis
from ._sparse_state_synthesis import cardinality_reduction
from ._ground_state_calibration import ground_state_calibration
from ._support_reduction import support_reduction
from ._qubit_decomposition import (
    select_pivot_qubit,
    to_controlled_gate,
    qubit_decomposition_opt,
)

EXACT_SYNTHESIS_QUBIT_THRESHOLD = 4
EXACT_SYNTHESIS_DENSITY_THRESHOLD = 10

EXACT_SYNTHESIS_CNOT_LIMIT = 10

ENABLE_EXACT_SYNTHESIS = True

ENABLE_QUBIT_REDUCTION = True
ENABLE_DENSITY_REDUCTION = True

ENABLE_DECOMPOSITION = False

ENABLE_REINDEX = True

ENABLE_PROGESS_BAR = True


class HybridCnotSynthesisStatistics:
    """Classes the hybridCnothesisStatistics class ."""

    def __init__(self) -> None:
        self.time_total: float = 0
        self.num_runs_support_reduction: int = 0
        self.time_support_reduction: float = 0
        self.time_exact_cnot_synthesis: float = 0
        self.time_density_reduction: float = 0
        self.time_qubit_decomposition: float = 0
        self.num_reduced_supports: int = 0
        self.num_reduced_density: int = 0
        self.num_saved_gates_decision: int = 0
        self.num_methods: dict = {}

    def report(self):
        """Report the number of runs supported by the benchmark ."""
        print("-" * 80)
        print(f"time_total: {self.time_total}")
        print("-" * 80)
        print(f"num_runs_support_reduction: {self.num_runs_support_reduction}")
        print(f"time_support_reduction: {self.time_support_reduction}")
        print(f"time_exact_cnot_synthesis: {self.time_exact_cnot_synthesis}")
        print(f"time_density_reduction: {self.time_density_reduction}")
        print(f"time_qubit_decomposition: {self.time_qubit_decomposition}")
        print(f"num_reduced_supports: {self.num_reduced_supports}")
        print(f"num_reduced_density: {self.num_reduced_density}")
        print(f"num_saved_gates_decision: {self.num_saved_gates_decision}")
        print("-" * 80)
        for method, num in self.num_methods.items():
            print(f"{method}: {num}")
        print("-" * 80)


def _hybrid_cnot_synthesis_impl(
    circuit: QCircuit,
    state: QState,
    verbose_level: int = 0,
    stats: HybridCnotSynthesisStatistics = None,
):
    prev_supports = state.get_supports()
    prev_num_supports = len(prev_supports)
    prev_density = state.get_sparsity()

    # first, run support reduction
    with stopwatch("support_reduction") as timer:
        state, support_reducing_gates = support_reduction(
            circuit, state, enable_cnot=True
        )

    num_cx_support_reduction = sum(
        [gate.get_cnot_cost() for gate in support_reducing_gates]
    )

    if ENABLE_REINDEX:
        # compact the state
        # now we get the sub state and sub circuit:
        sub_index_to_weight = {}
        old_to_new_qubit_mapping = {}

        supports = state.get_supports()
        num_supports = len(supports)
        for new_index, old_index in enumerate(supports):
            old_to_new_qubit_mapping[old_index] = new_index
        for index, weight in state.index_to_weight.items():
            new_index: int = 0
            for i, support in enumerate(supports):
                if index & (1 << support) != 0:
                    new_index |= 1 << i
            sub_index_to_weight[new_index] = weight
        state = QState(sub_index_to_weight, num_supports)
        circuit = circuit.sub_circuit(supports)

        # then we updates the supports to the new index
        supports = list(range(num_supports))

    # get the states
    supports = state.get_supports()
    num_supports = len(supports)
    density = state.get_sparsity()

    if ENABLE_PROGESS_BAR:
        print(f"num_supports: {num_supports:5d}, density: {density:5d}", end="\r")

    if stats is not None:
        stats.num_runs_support_reduction += 1
        stats.time_support_reduction += timer.time()
        stats.num_reduced_supports += prev_num_supports - num_supports
        stats.num_reduced_density += prev_density - density

    # check for the trivial case
    if density == 1:
        ground_state_calibration_gates = ground_state_calibration(circuit, state)

        gates = []
        for gate in ground_state_calibration_gates:
            gates.append(gate)
        for gate in support_reducing_gates:
            gates.append(gate)

        # ground state calibration has 0 CNOT
        return gates, num_cx_support_reduction

    # exact synthesis
    if (
        ENABLE_EXACT_SYNTHESIS
        and num_supports <= EXACT_SYNTHESIS_QUBIT_THRESHOLD
        and density <= EXACT_SYNTHESIS_DENSITY_THRESHOLD
    ):
        try:
            with stopwatch("exact_cnot_synthesis") as timer:
                exact_gates = exact_cnot_synthesis(
                    circuit,
                    state,
                    optimality_level=3,
                    verbose_level=verbose_level,
                    cnot_limit=EXACT_SYNTHESIS_CNOT_LIMIT,
                )
            if stats is not None:
                stats.time_exact_cnot_synthesis += timer.time()

            gates = []
            for gate in exact_gates:
                gates.append(gate)
            for gate in support_reducing_gates:
                gates.append(gate)

            num_cx_exact = sum([gate.get_cnot_cost() for gate in exact_gates])

            return gates, num_cx_exact
        except ValueError:
            pass

    # sparse state synthesis
    sparse_qsp_gates: List[QGate] = None
    num_sparse_qsp_cx: int = 0
    if ENABLE_DENSITY_REDUCTION:
        with stopwatch("cardinality_reduction") as timer:
            new_state, density_reduction_gates = cardinality_reduction(
                circuit, state, verbose_level=0
            )
        num_density_reduction_cx = sum(
            [gate.get_cnot_cost() for gate in density_reduction_gates]
        )

        if stats is not None:
            stats.time_density_reduction += timer.time()

        rec_gates, rec_cx = _hybrid_cnot_synthesis_impl(
            circuit,
            new_state,
            stats=stats,
        )

        sparse_qsp_gates = []
        for gate in rec_gates:
            sparse_qsp_gates.append(gate)
        for gate in density_reduction_gates:
            sparse_qsp_gates.append(gate)
        for gate in support_reducing_gates:
            sparse_qsp_gates.append(gate)

        num_sparse_qsp_cx = rec_cx + num_density_reduction_cx + num_cx_support_reduction

    # qubit decomposition
    qubit_reduction_gates: List[QGate] = None
    num_qubit_reduction_cx: int = 0
    if ENABLE_QUBIT_REDUCTION:
        with stopwatch("qubit_decomposition") as timer:
            qubit_decomposition_gates, new_state = qubit_decomposition_opt(
                circuit, state, supports
            )
        num_qubit_reduction_cx = sum(
            [gate.get_cnot_cost() for gate in qubit_decomposition_gates]
        )

        if stats is not None:
            stats.time_qubit_decomposition += timer.time()

        rec_gates, rec_cx = _hybrid_cnot_synthesis_impl(
            circuit,
            new_state,
            stats=stats,
        )

        qubit_reduction_gates = []
        for gate in rec_gates:
            qubit_reduction_gates.append(gate)
        for gate in qubit_decomposition_gates:
            qubit_reduction_gates.append(gate)
        for gate in support_reducing_gates:
            qubit_reduction_gates.append(gate)

        num_qubit_reduction_cx += rec_cx + num_cx_support_reduction

    qubit_decomposition_gates: List[QGate] = None
    num_qubit_decomposition_cx: int = 0  # not implemented yet
    if ENABLE_DECOMPOSITION:
        pivot = select_pivot_qubit(state, supports)
        pivot_qubit = circuit.qubit_at(pivot)

        neg_state, pos_state, weights0, weights1 = state.cofactors(pivot)

        # we first add a rotation gate to the pivot qubit
        theta = 2 * np.arctan(weights1 / weights0)
        ry_gate = RY(theta, pivot_qubit)

        pos_gates = _hybrid_cnot_synthesis_impl(
            circuit,
            pos_state,
            stats=stats,
        )

        neg_gates = _hybrid_cnot_synthesis_impl(
            circuit,
            neg_state,
            stats=stats,
        )

        qubit_decomposition_gates = []
        qubit_decomposition_gates.append(ry_gate)
        for gate in pos_gates:
            controlled_gate = to_controlled_gate(gate, pivot_qubit, True)
            qubit_decomposition_gates.append(controlled_gate)

        for gate in neg_gates:
            controlled_gate = to_controlled_gate(gate, pivot_qubit, False)
            qubit_decomposition_gates.append(controlled_gate)

        for gate in support_reducing_gates:
            qubit_decomposition_gates.append(gate)

    # we choose the best one
    # based on the number of CNOT gates
    Method = namedtuple("method", ["name", "gates", "num_gates"])

    candidates = []

    if sparse_qsp_gates is not None:
        candidates.append(Method("sparse_qsp", sparse_qsp_gates, num_sparse_qsp_cx))
    if qubit_reduction_gates is not None:
        candidates.append(
            Method("qubit_reduction", qubit_reduction_gates, num_qubit_reduction_cx)
        )
    if qubit_decomposition_gates is not None:
        candidates.append(
            Method(
                "qubit_decomposition",
                qubit_decomposition_gates,
                num_qubit_decomposition_cx,
            )
        )

    # pylint: disable=unnecessary-lambda
    best_candidate = min(candidates, key=lambda x: x.num_gates)
    worst_candidate = max(candidates, key=lambda x: x.num_gates)

    best_gates = best_candidate.gates
    best_method = best_candidate.name

    worst_num_gates = worst_candidate.num_gates
    best_num_gates = best_candidate.num_gates

    if stats is not None:
        stats.num_saved_gates_decision += worst_num_gates - best_num_gates
        stats.num_methods[best_method] = stats.num_methods.get(best_method, 0) + 1

    return best_gates, best_num_gates


def hybrid_cnot_synthesis(
    state: QState,
    map_gates: bool = True,
    verbose_level: int = 0,
    stats: HybridCnotSynthesisStatistics = None,
):
    """A hybrid method combining both qubit- and cardinality- reduction.
    The solver would choose the best method based on a Markov decision process.

    :param state: [description]
    :type state: QState
    :param map_gates: [description], defaults to False
    :type map_gates: bool, optional
    :return: [description]
    :rtype: [type]
    """

    # check the initial state
    num_qubits = state.num_qubits
    density = state.get_sparsity()

    # 0.6 is a magic number,
    density_reduction_cnot_estimation = int(density * num_qubits)
    qubit_reduction_cnot_estimation = 1 << num_qubits

    # pylint: disable=W0603

    global ENABLE_QUBIT_REDUCTION
    global ENABLE_DENSITY_REDUCTION
    if density_reduction_cnot_estimation < qubit_reduction_cnot_estimation:
        print_yellow("ENABLE_DENSITY_REDUCTION")
        ENABLE_QUBIT_REDUCTION = False
        ENABLE_DENSITY_REDUCTION = True
    else:
        print_yellow("ENABLE_QUBIT_REDUCTION")
        ENABLE_QUBIT_REDUCTION = True
        ENABLE_DENSITY_REDUCTION = False

    circuit = QCircuit(state.num_qubits, map_gates=map_gates)

    with stopwatch("hybrid_cnot_synthesis") as timer:
        gates, _ = _hybrid_cnot_synthesis_impl(
            circuit,
            state,
            verbose_level=verbose_level,
            stats=stats,
        )

    if stats is not None:
        stats.time_total = timer.time()

    circuit.add_gates(gates)

    global_stopwatch_report()
    return circuit



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-20 00:05:16
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-20 00:06:35
"""

import numpy as np


def C_state(num_qubits: int, threshold: int) -> np.ndarray:
    """Return the C state corresponding to the given number of qubits and the given threshold .

    :param num_qubits: [description]
    :type num_qubits: int
    :param threshold: [description]
    :type threshold: int
    :return: [description]
    :rtype: np.ndarray
    """
    state = np.zeros(2**num_qubits)

    ones: float = 0

    for i in range(2**num_qubits):
        if i < threshold:
            state[i] = 1
            ones += 1

    return state / np.sqrt(ones)



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-11-21 12:51:01
Last Modified by: Hanyu Wang
Last Modified time: 2023-11-21 12:59:21
"""

from typing import List

import numpy as np


def get_ry_angles(self, qubit_index: int) -> List[float]:
    """Return the projection of the state .

    :param qubit_index: [description]
    :type qubit_index: int
    :return: [description]
    :rtype: List[float]
    """
    thetas = []
    for idx in self.index_set:
        idx0 = idx & ~(1 << qubit_index)
        idx1 = idx0 ^ (1 << qubit_index)

        # check if the qubit is 1
        if idx == idx1:
            if idx0 not in self.index_set:
                thetas.append(np.pi)
            continue

        if idx1 not in self.index_set:
            thetas.append(0)
            continue

        # now we check the rotation angle
        weight_from = self.index_to_weight[idx0]
        weight_total = np.sqrt(
            (self.index_to_weight[idx1] ** 2) + (self.index_to_weight[idx0] ** 2)
        )

        if self.index_to_weight[idx1] > 0:
            _theta = 2 * np.arccos(weight_from / weight_total)
        else:
            _theta = -2 * np.arccos(weight_from / weight_total)

        thetas.append(_theta)

    return thetas


def get_rotation_table(self, qubit_index: int) -> List[float]:
    """Return the projection of the state .

    :param qubit_index: [description]
    :type qubit_index: int
    :return: [description]
    :rtype: List[float]
    """
    thetas = {}
    for idx in self.index_set:
        idx0 = idx & ~(1 << qubit_index)
        idx1 = idx0 ^ (1 << qubit_index)

        # check if the qubit is 1
        if idx == idx1:
            if idx0 not in self.index_set:
                thetas[idx0] = np.pi
            continue

        if idx1 not in self.index_set:
            thetas[idx0] = 0
            continue

        # now we check the rotation angle
        weight_from = self.index_to_weight[idx0]
        weight_total = np.sqrt(
            (self.index_to_weight[idx1] ** 2) + (self.index_to_weight[idx0] ** 2)
        )

        if self.index_to_weight[idx1] > 0:
            _theta = 2 * np.arccos(weight_from / weight_total)
        else:
            _theta = -2 * np.arccos(weight_from / weight_total)

        thetas[idx0] = _theta

    return thetas


def get_cry_angles(
    self, control_qubit_index: int, target_qubit_index: int
) -> List[float]:
    """Return the projection of the state .

    :param qubit_index: [description]
    :type qubit_index: int
    :return: [description]
    :rtype: List[float]
    """
    thetas = {}
    for idx in self.index_set:
        idx0 = idx & ~(1 << target_qubit_index)
        idx1 = idx0 ^ (1 << target_qubit_index)

        # check if the qubit is 1
        if idx == idx1:
            if idx0 not in self.index_set:
                thetas[idx0] = np.pi
            continue

        if idx1 not in self.index_set:
            thetas[idx0] = 0
            continue

        # now we check the rotation angle
        weight_from = self.index_to_weight[idx0]
        weight_total = np.sqrt(
            (self.index_to_weight[idx1] ** 2) + (self.index_to_weight[idx0] ** 2)
        )

        if self.index_to_weight[idx1] > 0:
            _theta = 2 * np.arccos(weight_from / weight_total)
        else:
            _theta = -2 * np.arccos(weight_from / weight_total)

        thetas[idx0] = _theta

    cry_thetas = []
    for idx, theta in thetas.items():
        rdx = idx ^ (1 << control_qubit_index)

        if (idx >> control_qubit_index) & 1 == 1:
            continue

        if rdx not in thetas:
            continue

        if np.isclose(theta, thetas[rdx]):
            continue

        # if theta != 0 and thetas[rdx] != 0:
        #     continue

        # this is the axis to run reflection
        beta = (thetas[rdx] + theta) / 2

        # need to rotate to the |+> state to perform CX
        cry_thetas.append(np.pi / 2 - beta)

    return cry_thetas


def get_most_frequent_theta(self, qubit_index: int) -> float:
    """Return the projection of the state .

    :param qubit_index: [description]
    :type qubit_index: int
    :return: [description]
    :rtype: List[float]
    """
    thetas = self.get_ry_angles(qubit_index)

    best_theta = None
    best_theta_count = 0

    curr_theta = None
    curr_theta_count = 0
    for theta in sorted(thetas):
        if curr_theta is None:
            curr_theta = theta
            curr_theta_count = 1

            best_theta = curr_theta
            best_theta_count = curr_theta_count
        elif np.isclose(theta, curr_theta):
            curr_theta_count += 1
        else:
            if curr_theta_count > best_theta_count:
                best_theta = curr_theta
                best_theta_count = curr_theta_count
            curr_theta = theta
            curr_theta_count = 1

    return best_theta



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-11-21 21:50:06
Last Modified by: Hanyu Wang
Last Modified time: 2023-11-21 22:37:44
"""

from typing import List, Tuple

from queue import PriorityQueue

import xyz.circuit as qc
from xyz.qstate import QState


class Library:
    """A library class .

    a library is linked to a set of operators when initialized .
    """

    def __init__(self) -> None:
        self.visited_states = set()
        self.state_queue = PriorityQueue()
        self.enquened_states = {}
        self.record = {}
        self.num_qubits: int = None
        self.circuit: qc.QCircuit = None

    def explore_state(
        self, curr_state_pair: Tuple, next_state_pair: Tuple, gates: List
    ):
        """AI is creating summary for explore_state

        :param curr_state: [description]
        :type curr_state: QState
        """

        _, curr_state = curr_state_pair
        next_cost, next_state = next_state_pair

        next_state_repr = next_state.repr()

        # we skip the state if it is already visited
        if next_state_repr in self.visited_states:
            return None

        # we skip the state if it is already enquened and the cost is higher
        if (
            next_state_repr in self.enquened_states
            and next_cost >= self.enquened_states[next_state_repr]
        ):
            return None

        # now we add the state to the queue
        self.state_queue.put((next_cost, next_state))
        self.enquened_states[next_state_repr] = next_cost

        self.record[hash(next_state)] = hash(curr_state), gates[:]

    def initialize_queue(self, target_state: QState, circuit: qc.QCircuit):
        """Initialize the queue for the given state .

        :param target_state: [description]
        :type target_state: QState
        """

        self.visited_states = set()
        self.state_queue = PriorityQueue()
        self.enquened_states = {}
        self.record = {}
        self.state_queue.put((0, target_state))
        self.num_qubits = target_state.num_qubits
        self.circuit = circuit

    def explore(self):
        """This method is called when the client is opened ."""
        raise NotImplementedError

    def get_solution(self):
        """This method is called when the client is closed ."""

        gates = []
        backtraced_states: set = set()
        curr_hash = hash(QState.ground_state(self.num_qubits))

        assert curr_hash in self.record

        while curr_hash in self.record:
            if curr_hash in backtraced_states:
                raise ValueError("Loop found")
            backtraced_states.add(curr_hash)
            prev_hash, _gates = self.record[curr_hash]
            for _gate in _gates:
                gates.append(_gate)
            curr_hash = prev_hash

        return gates



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-09-08 12:59:35
Last Modified by: Hanyu Wang
Last Modified time: 2023-09-10 17:52:33
"""


import copy
import numpy as np


from xyz.circuit.basic_gates.cx import CX
from xyz.circuit.basic_gates.mcry import MCRY
from xyz.circuit.qcircuit import QCircuit

from xyz.qstate import QState
from ._ground_state_calibration import ground_state_calibration


def _maximize_difference_once(state: QState, indices: set, diff_lit: dict):
    max_difference: int = -1

    best_index_set_0 = set()
    best_qubit = None
    best_value = None

    assert len(indices) >= 2
    length = len(indices)

    for qubit in range(state.num_qubits):
        # we cannot be better than this
        if max_difference == length - 1:
            break

        if qubit in diff_lit:
            continue

        index_set_0 = set()

        for index in indices:
            if index & (1 << qubit) == 0:
                index_set_0.add(index)

        length0 = len(index_set_0)
        difference = abs(length - (length0 << 1))

        # this means either the qubit is all 0 or all 1
        if difference == length:
            continue

        # we update the best difference
        if difference > max_difference:
            max_difference = difference
            best_index_set_0 = index_set_0
            best_qubit = qubit

            # this will return 0 if qubit has less 0
            # this will return 1 if qubit has more 0
            best_value = bool(length < (1 << length0))

    assert best_qubit is not None
    assert len(best_index_set_0) != 0 and len(best_index_set_0) != length

    best_index_set = best_index_set_0 if best_value == 0 else indices - best_index_set_0

    return best_index_set, best_qubit, best_value


def cardinality_reduction(circuit: QCircuit, state: QState, verbose_level: int = 0):
    """Select indices for the QState .

    :param state: [description]
    :type state: QState
    """

    diff_qubits = []
    diff_values = []

    # we first select the indices that maximize the difference
    diff_lits = []

    indices = set(list(state.index_set)[:])
    while len(indices) > 1:
        index_set, qubit, value = _maximize_difference_once(state, indices, diff_lits)

        diff_lits.append((qubit, value))

        indices = index_set

    assert len(indices) == 1

    index1 = list(indices)[0]

    if verbose_level >= 3:
        print(f"indices: {indices}, diff_lits: {diff_lits}")

    # we will later use diff_qubit and diff_value as CNOTs
    diff_qubit, diff_value = diff_lits.pop()

    # now we select the second index
    index2_candidates = set()
    for index in set(state.index_set) - indices:
        violated: bool = False
        for _diff_qubit, _diff_value in diff_lits:
            if (index >> _diff_qubit) & 1 != _diff_value:
                violated = True
                break
        if violated:
            continue
        # we add the index to the candidates
        index2_candidates.add(index)

    # we select the second index
    indices = set(list(index2_candidates)[:])
    while len(indices) > 1:
        index_set, qubit, value = _maximize_difference_once(state, indices, diff_lits)

        diff_lits.append((qubit, value))
        indices = index_set

    assert (
        len(indices) == 1
    ), f"indices: {indices}, index2_candidates: {index2_candidates}, diff_lits: {diff_lits}"

    index2 = list(indices)[0]

    if verbose_level >= 3:
        print(f"indices: {indices}, diff_lits: {diff_lits}")

    # now we try to merge the two states, by flipping the index1 to approach index2
    if verbose_level >= 3:
        print(f"merging indices from {index1} to {index2}")
    gates = []

    index_to_weight = copy.copy(state.index_to_weight)

    for qubit in range(state.num_qubits):
        if (index1 >> qubit) & 1 == (index2 >> qubit) & 1:
            continue

        if qubit == diff_qubit:
            continue

        control_qubit = circuit.qubit_at(diff_qubit)
        control_phase = diff_value
        target_qubit = circuit.qubit_at(qubit)

        gate = CX(control_qubit, control_phase, target_qubit)
        gates.append(gate)
        if verbose_level >= 3:
            print(f"\t adding gate {gate} to the circuit")

        # modify the index_to_weight according to the CNOT
        new_index_to_weight = {}
        for index, weight in index_to_weight.items():
            reserved_index = index ^ (1 << qubit)
            if (index >> diff_qubit) & 1 == diff_value:
                new_index_to_weight[reserved_index] = weight
            else:
                new_index_to_weight[index] = weight

        index_to_weight = new_index_to_weight

    diff_qubits, diff_values = [], []
    if len(diff_lits) > 0:
        diff_qubits, diff_values = zip(*diff_lits)

    control_qubits = [circuit.qubit_at(qubit) for qubit in diff_qubits]
    control_phases = diff_values
    target_qubit = circuit.qubit_at(diff_qubit)

    assert index2 in index_to_weight
    reversed_index2 = index2 ^ (1 << diff_qubit)
    if verbose_level >= 3:
        print(
            f"index2 = {index2}, reversed_index2: {reversed_index2}, index_to_weight: {index_to_weight}"
        )
    assert reversed_index2 in index_to_weight

    theta = 2 * np.arctan(index_to_weight[reversed_index2] / index_to_weight[index2])

    gates.append(MCRY(theta, control_qubits, control_phases, target_qubit))

    index_to_weight[index2] = np.sqrt(
        index_to_weight[index2] ** 2 + index_to_weight[reversed_index2] ** 2
    )

    index_to_weight.pop(reversed_index2)

    # Update the state
    new_state = QState(index_to_weight, state.num_qubits)

    return new_state, gates[::-1]


def sparse_state_synthesis(circuit, state: QState, verbose_level: int = 0):
    """This function is used to synthesis sparse state.

    reference: https://github.com/qclib/qclib/blob/master/qclib/state_preparation/merge.py

    """

    # deep copy
    curr_state = QState(state.index_to_weight, state.num_qubits)

    gates = []

    while True:
        density = len(curr_state.index_set)

        if verbose_level >= 3:
            print(f"Current state: {curr_state}, density: {density}")

        # we reach the end of the state
        if density == 1:
            break

        new_state, _gates = cardinality_reduction(
            circuit, curr_state, verbose_level=verbose_level
        )
        for gate in _gates:
            gates.append(gate)

        curr_state = new_state

    preprocessing_gates = ground_state_calibration(circuit, curr_state)

    return preprocessing_gates + gates[::-1]



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-11-06 12:22:56
Last Modified by: Hanyu Wang
Last Modified time: 2023-11-06 12:41:58
"""

from setuptools import setup, find_packages

setup(
    name="quantum-xyz",
    version="0.1.0",
    author="Hanyu Wang",
    description="A quantum circuit synthesis library",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(include=["xyz", "xyz.*"]),
    install_requires=[
        """
            llist==0.7.1
            matplotlib==3.7.1
            numpy==1.23.5
            qiskit==0.43.1
            qiskit_ibmq_provider==0.20.2
            qiskit_terra==0.24.1
            scipy==1.10.1
        """
    ],
)



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-18 11:29:54
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-18 11:33:28
"""

from .synthesis import *
from .verification import *



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-20 18:44:02
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-22 23:55:41
"""

from enum import Enum, auto

from .qubit import QBit


class QGateType(Enum):
    """This method is used to set the auto - gate type for QGate type .

    When defining a new gate, we need to register here to get the type name of the gate .

    :param Enum: [description]
    :type Enum: [type]
    """

    U = auto()
    CU = auto()
    MCU = auto()

    X = auto()
    Y = auto()
    Z = auto()

    CY = auto()
    CZ = auto()
    CX = auto()

    # single qubit rotation gates
    RX = auto()
    RY = auto()
    RZ = auto()

    # controlled rotation gates
    CRX = auto()
    CRY = auto()
    CRZ = auto()

    MCRY = auto()

    MULTIPLEX_Y = auto()

    MCMY = auto()

    NONE = auto()


class QGate:
    """Class method for creating a new gate class ."""

    def __init__(self, qgate_type: QGateType) -> None:
        self.qgate_type = qgate_type

    def __str__(self) -> str:
        return self.qgate_type.name

    def get_qgate_type(self) -> QGateType:
        """Returns the q gate type .

        :return: [description]
        :rtype: QGateType
        """
        return self.qgate_type


class BasicGate(QGate):
    """Class method for creating a gate .

    :param QGate: [description]
    :type QGate: [type]
    """

    def __init__(self, qgate_type: QGateType, target_qubit: QBit) -> None:
        QGate.__init__(self, qgate_type)

        if not isinstance(target_qubit, QBit):
            print(f"WARNING: target_qubit {target_qubit} is not a QBit")

        assert isinstance(target_qubit, QBit)
        self.target_qubit: QBit = target_qubit

    def get_target_qubit(self) -> QBit:
        """Returns the target qubit .

        :return: [description]
        :rtype: QBit
        """
        return self.target_qubit


class AdvancedGate(QGate):
    """Class method that creates a AdvancedGate class .

    :param QGate: [description]
    :type QGate: [type]
    """

    def __init__(self, qgate_type: QGateType) -> None:
        QGate.__init__(self, qgate_type)



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-22 13:11:53
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-22 23:32:05
"""

from .base import BasicGate, ControlledGate, QBit, QGateType, RotationGate


class CRY(RotationGate, BasicGate, ControlledGate):
    """Classmethod to create CRY and ControlledGate gate .

    :param RotationGate: [description]
    :type RotationGate: [type]
    :param BasicGate: [description]
    :type BasicGate: [type]
    :param ControlledGate: [description]
    :type ControlledGate: [type]
    """

    def __init__(
        self, theta: float, control_qubit: QBit, phase: int, target_qubit: QBit
    ) -> None:
        BasicGate.__init__(self, QGateType.CRY, target_qubit)
        RotationGate.__init__(self, theta)
        ControlledGate.__init__(self, control_qubit, phase)

    def __str__(self) -> str:
        return f"CRY({self.theta:0.02f}, {self.control_qubit}[{self.phase}])"

    def get_cnot_cost(self) -> int:
        """Returns the cost of the cost of the gate.

        :return: [description]
        :rtype: int
        """
        return 2



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-09-01 12:51:03
Last Modified by: Hanyu Wang
Last Modified time: 2023-09-01 12:56:17
"""

from queue import PriorityQueue
from xyz.circuit.basic_gates.base.gate import QGate
from xyz.circuit.basic_gates.cx import CX
from xyz.circuit import QCircuit, X, MCRY

import xyz.qstate as qs
import xyz.operator as op

from ._ground_state_calibration import ground_state_calibration


class AStarCost:
    """AStarCost class ."""

    def __init__(
        self, cnot_cost: float, unitary_cost: float, lower_bound: float
    ) -> None:
        self.cnot_cost = cnot_cost
        self.unitary_cost = unitary_cost
        self.lower_bound = lower_bound

    def __lt__(self, other):
        return (
            self.cnot_cost + 0.1 * self.unitary_cost + self.lower_bound
            < other.cnot_cost + 0.1 * other.unitary_cost + other.lower_bound
        )

    def __ge__(self, other):
        return (
            self.cnot_cost + 0.1 * self.unitary_cost + self.lower_bound
            >= other.cnot_cost + 0.1 * other.unitary_cost + other.lower_bound
        )

    def __str__(self) -> str:
        return f"cnot_cost: {self.cnot_cost}, unitary_cost: {self.unitary_cost}, lower_bound: {self.lower_bound}"


def exact_cnot_synthesis(
    circuit: QCircuit,
    target_state: qs.QState,
    optimality_level: int = 3,
    verbose_level: int = 0,
    cnot_limit: int = None,
):
    """This function finds the exact cnot_cnot_synthesis of a circuit .

    :param circuit: [description]
    :type circuit: QCircuit
    :param qubit_mapping: [description]
    :type qubit_mapping: dict
    :param target_state: [description]
    :type target_state: qs.QState
    :param optimality_level: [description], defaults to 3
    :type optimality_level: int, optional
    :param verbose_level: [description], defaults to 0
    :type verbose_level: int, optional
    :raises ValueError: [description]
    :return: [description]
    :rtype: [type]
    """

    # now we start the search
    visited_states = set()
    state_queue = PriorityQueue()
    enquened_states = {}
    record = {}

    def map_qubit(qubit_index: int) -> int:
        return circuit.qubit_at(qubit_index)

    def explore_state(
        curr_state: qs.QState, quantum_operator: op.QOperator, curr_cost: AStarCost
    ) -> qs.QState:
        """Explore a state in a SRGraph ."""
        nonlocal visited_states, state_queue, enquened_states, record
        try:
            next_state = quantum_operator(curr_state)
            if verbose_level >= 3:
                print(
                    f"next_state: {next_state}, curr_state = {curr_state} gate = {quantum_operator}"
                )
        except ValueError:
            return None

        next_cost = AStarCost(
            curr_cost.cnot_cost + quantum_operator.get_cost(),
            curr_cost.unitary_cost + 1,
            next_state.get_lower_bound(),
        )

        next_state_repr = next_state.repr()

        # we skip the state if it is already visited
        if next_state_repr in visited_states:
            return None

        # we skip the state if it is already enquened and the cost is higher
        if (
            next_state_repr in enquened_states
            and next_cost >= enquened_states[next_state_repr]
        ):
            return None

        # now we add the state to the queue
        state_queue.put((next_cost, next_state))
        enquened_states[next_state_repr] = next_cost

        # we record the gate
        gate_to_record: QGate = None
        match quantum_operator.operator_type:
            case op.QOperatorType.X:
                gate_to_record = X(map_qubit(quantum_operator.target_qubit_index))
            case op.QOperatorType.CX:
                gate_to_record = CX(
                    map_qubit(quantum_operator.control_qubit_index),
                    quantum_operator.control_qubit_phase,
                    map_qubit(quantum_operator.target_qubit_index),
                )
            case op.QOperatorType.T0 | op.QOperatorType.T1:
                gate_to_record = MCRY(
                    quantum_operator.theta,
                    [],
                    [],
                    map_qubit(quantum_operator.target_qubit_index),
                )

            case op.QOperatorType.CT0 | op.QOperatorType.CT1:
                gate_to_record = MCRY(
                    quantum_operator.theta,
                    [map_qubit(quantum_operator.control_qubit_index)],
                    [quantum_operator.control_qubit_phase],
                    map_qubit(quantum_operator.target_qubit_index),
                )

        # and record the quantum_operator
        if verbose_level >= 3:
            print(
                f"recording [{hash(next_state)}] <- {hash(curr_state)}, gate: {gate_to_record}"
            )
        record[hash(next_state)] = hash(curr_state), gate_to_record

        return next_state

    # begin of the exact synthesis algorithm
    num_qubits = target_state.num_qubits
    initial_state = qs.QState.ground_state(num_qubits)

    curr_state = target_state
    curr_cost = AStarCost(0, 0, curr_state.get_lower_bound())
    state_queue.put((curr_cost, curr_state))

    solution_reached: bool = False

    sparsity = target_state.get_sparsity()
    best_supports = target_state.get_supports()

    # This function is called by the search loop.
    while not state_queue.empty():
        curr_cost, curr_state = state_queue.get()

        if verbose_level >= 2:
            print(f"\n\ncurr_state: {curr_state}, cost: {curr_cost}")

        if cnot_limit is not None and curr_cost.cnot_cost > cnot_limit:
            # this will then raise an ValueError
            break

        if curr_state == initial_state:
            # then we have found the solution
            solution_reached = True
            break

        _sparsity = curr_state.get_sparsity()
        _supports = curr_state.get_supports()

        _optimality_level = optimality_level

        if _sparsity == 1:
            solution_reached = True
            break

        if _sparsity >= 2 and _sparsity < sparsity:
            sparsity = _sparsity
            if verbose_level >= 1:
                print(f"sparsity: {sparsity}")

            if _optimality_level <= 2:
                state_queue = PriorityQueue()
                enquened_states = {}

        if len(_supports) < len(best_supports):
            best_supports = _supports[:]
            if verbose_level >= 1:
                print(f"supports: {_supports}, length = {len(_supports)}")

            if _optimality_level <= 2:
                state_queue = PriorityQueue()
                enquened_states = {}

        curr_state_repr = curr_state.repr()
        visited_states.add(curr_state_repr)
        supports = curr_state.get_supports()

        search_done = False

        # try dependency analysis
        if not search_done:
            signatures = curr_state.get_qubit_signatures()
            const1 = curr_state.get_const1_signature()
            signature_to_qubits = {}
            for qubit_index, signature in enumerate(signatures):
                if signature == 0 or signature == const1:
                    continue
                if signature in signature_to_qubits:
                    quantum_operator = op.CXOperator(
                        qubit_index, signature_to_qubits[signature], True
                    )
                    next_state = explore_state(curr_state, quantum_operator, curr_cost)
                    search_done = True
                    break
                elif signature ^ const1 in signature_to_qubits:
                    quantum_operator = op.CXOperator(
                        qubit_index, signature_to_qubits[signature ^ const1], False
                    )
                    next_state = explore_state(curr_state, quantum_operator, curr_cost)
                    search_done = True
                    break
                else:
                    signature_to_qubits[signature] = qubit_index

            # try higher level dependency analysis
            if _optimality_level <= 2:
                for qubit_index1, signature1 in enumerate(signatures):
                    if signature1 == 0 or signature1 == const1:
                        continue
                    for qubit_index2, signature2 in enumerate(signatures):
                        if qubit_index1 == qubit_index2:
                            continue
                        if signature2 == 0 or signature2 == const1:
                            continue

                        if signature1 ^ signature2 in signature_to_qubits:
                            quantum_operator = op.CXOperator(
                                qubit_index1,
                                signature_to_qubits[signature1 ^ signature2],
                                True,
                            )
                            next_state = explore_state(
                                curr_state, quantum_operator, curr_cost
                            )
                            if next_state is not None:
                                quantum_operator = op.CXOperator(
                                    qubit_index2,
                                    signature_to_qubits[signature1 ^ signature2],
                                    True,
                                )
                                explore_state(next_state, quantum_operator, curr_cost)

                            search_done = True
                            break

                        elif signature1 ^ signature2 ^ const1 in signature_to_qubits:
                            quantum_operator = op.CXOperator(
                                qubit_index1,
                                signature_to_qubits[signature1 ^ signature2 ^ const1],
                                True,
                            )
                            next_state = explore_state(
                                curr_state, quantum_operator, curr_cost
                            )
                            if next_state is not None:
                                quantum_operator = op.CXOperator(
                                    qubit_index2,
                                    signature_to_qubits[
                                        signature1 ^ signature2 ^ const1
                                    ],
                                    False,
                                )
                                explore_state(next_state, quantum_operator, curr_cost)

                            search_done = True
                            break
                        else:
                            pass

        # apply merge0
        if not search_done:
            for target_qubit in supports:
                quantum_operator = op.TROperator(target_qubit, True)
                explore_state(curr_state, quantum_operator, curr_cost)
                quantum_operator = op.TROperator(target_qubit, False)
                explore_state(curr_state, quantum_operator, curr_cost)

        # apply cmerge
        if not search_done:
            for target_qubit in supports:
                for control_qubit in supports:
                    if control_qubit == target_qubit:
                        continue
                    for phase in [True, False]:
                        for target_phase in [True, False]:
                            quantum_operator = op.CTROperator(
                                target_qubit, target_phase, control_qubit, phase
                            )
                            explore_state(curr_state, quantum_operator, curr_cost)

        # CNOT
        if not search_done:
            for target_qubit in supports:
                for control_qubit in supports:
                    if control_qubit == target_qubit:
                        continue
                    for phase in [True, False]:
                        quantum_operator = op.CXOperator(
                            target_qubit, control_qubit, phase
                        )
                        explore_state(curr_state, quantum_operator, curr_cost)

        # apply x
        if not search_done:
            if _optimality_level <= 3 and curr_state.get_sparsity() != 1:
                # we can prove that we only need to apply X once, so we skip the rest
                pass
            else:
                for target_qubit in supports:
                    quantum_operator = op.XOperator(target_qubit)
                    explore_state(curr_state, quantum_operator, curr_cost)

    if not solution_reached:
        raise ValueError("No solution found")

    final_state = qs.QState(curr_state.index_to_weight, curr_state.num_qubits)

    x_gates = ground_state_calibration(circuit, final_state)

    if verbose_level >= 2:
        print("\n\n")

    for record_key, record_value in record.items():
        prev_state, gate = record_value
        if verbose_level >= 2:
            print(
                f"record_key: {record_key}\n\t prev_state: {prev_state}\n\t gate: {gate}"
            )

    gates = x_gates[:]
    backtraced_states: set = set()
    curr_hash = hash(curr_state)
    while curr_hash in record:
        if curr_hash in backtraced_states:
            raise ValueError("Loop found")
        backtraced_states.add(curr_hash)
        prev_hash, gate = record[curr_hash]
        gates.append(gate)
        curr_hash = prev_hash
        if verbose_level >= 2:
            print(f"curr_hash: {curr_hash:0b}, gate: {gate}")

    return gates



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-09-01 12:48:05
Last Modified by: Hanyu Wang
Last Modified time: 2023-09-01 13:05:14
"""

import threading
from typing import List
import numpy as np

from xyz.circuit import QGate, QGateType, QBit, CX, CRY
from xyz.circuit.basic_gates.mcmy import MCMY
from xyz.circuit.basic_gates.mcry import MCRY
from xyz.circuit.basic_gates.ry import RY

from xyz.circuit.qcircuit import QCircuit
from xyz.qstate import QState
from ._exact_cnot_synthesis import exact_cnot_synthesis
from ._support_reduction import support_reduction


def to_controlled_gate(gate: QGate, control_qubit: QBit, control_phase: bool):
    """Return a controlled gate .

    :param gate: [description]
    :type gate: QGate
    :return: [description]
    :rtype: [type]
    """
    assert control_qubit != gate.get_target_qubit()
    match gate.get_qgate_type():
        case QGateType.X:
            return CX(control_qubit, control_phase, gate.get_target_qubit())
        case QGateType.RY:
            return CRY(
                gate.get_theta(), control_qubit, control_phase, gate.get_target_qubit()
            )
        case QGateType.CX:
            control_qubits = [control_qubit, gate.get_control_qubit()]
            phases = [control_phase, gate.get_phase()]
            return MCRY(np.pi, control_qubits, phases, gate.get_target_qubit())
        case QGateType.CRY:
            control_qubits = [control_qubit, gate.get_control_qubit()]
            phases = [control_phase, gate.get_phase()]
            return MCRY(
                gate.get_theta(), control_qubits, phases, gate.get_target_qubit()
            )

        case QGateType.MCRY:
            control_qubits = [control_qubit] + gate.get_control_qubits()
            phases = [control_phase] + gate.get_phases()
            return MCRY(
                gate.get_theta(), control_qubits, phases, gate.get_target_qubit()
            )

        case _:
            raise NotImplementedError(
                f"Controlled gate {gate.get_qgate_type()} is not implemented"
            )


def select_pivot_qubit(state: QState, supports: set):
    """Selects the pivot of the given state .

    :param state: [description]
    :type state: QState
    :param supports: [description]
    :type supports: set
    :return: [description]
    :rtype: [type]
    """
    max_difference: int = -1
    best_qubit = None

    indices = state.index_set

    assert (
        len(indices) >= 2
    ), f"state = {state}, len(indices) = {len(indices)}, supports = {supports}"
    length = len(indices)

    for qubit in supports:
        # we cannot be better than this
        if max_difference == length - 1:
            break

        index_set_0 = set()

        for index in indices:
            if index & (1 << qubit) == 0:
                index_set_0.add(index)

        length0 = len(index_set_0)
        difference = abs(length - (length0 << 1))

        # we update the best difference
        if difference > max_difference:
            max_difference = difference
            best_qubit = qubit

    assert best_qubit is not None
    return best_qubit


def select_informative_qubit(state: QState, supports: set):
    """Selects the smallest qubit in the given state .

    :param state: [description]
    :type state: QState
    :param supports: [description]
    :type supports: set
    :return: [description]
    :rtype: [type]
    """
    best_qubit = None

    indices = state.index_set

    assert (
        len(indices) >= 2
    ), f"state = {state}, len(indices) = {len(indices)}, supports = {supports}"
    length = len(indices)
    min_difference: int = length + 1

    for qubit in supports:
        index_set_0 = set()

        for index in indices:
            if index & (1 << qubit) == 0:
                index_set_0.add(index)

        length0 = len(index_set_0)
        difference = abs(length - (length0 << 1))

        # we update the best difference
        if difference < min_difference:
            min_difference = difference
            best_qubit = qubit

    assert best_qubit is not None
    return best_qubit


def _qubit_decomposition_impl(
    circuit: QCircuit,
    gates: List[QGate],
    state: QState,
    optimality_level: int = 3,
    multi_thread: bool = False,
    verbose_level: int = 0,
    cnot_limit: int = None,
):
    assert len(gates) == 0

    # we first run qubit reduction
    state, support_gates = support_reduction(circuit, state)

    supports = state.get_supports()
    num_supports = len(supports)

    if num_supports <= 4:
        # we can use optimality_level=3
        exact_gates = exact_cnot_synthesis(
            circuit,
            state,
            optimality_level=3,
            verbose_level=verbose_level,
            cnot_limit=cnot_limit,
        )
        for gate in exact_gates:
            gates.append(gate)

        for gate in support_gates:
            gates.append(gate)
        return

    # we select the qubits that maximize the entropy
    pivot = select_pivot_qubit(state, supports)
    pivot_qubit = circuit.qubit_at(pivot)

    neg_state, pos_state, weights0, weights1 = state.cofactors(pivot)

    # we first add a rotation_entry gate to the pivot qubit
    theta = 2 * np.arctan(weights1 / weights0)
    gate = RY(theta, pivot_qubit)
    gates.append(gate)

    # then we recursively decompose the two substates
    pos_gates = []
    neg_gates = []

    # using multi_thread, we can parallelize the decomposition
    if multi_thread:
        thread_pos = threading.Thread(
            target=_qubit_decomposition_impl,
            args=(
                circuit,
                pos_gates,
                pos_state,
                optimality_level,
                multi_thread,
                verbose_level,
                cnot_limit,
            ),
        )
        thread_neg = threading.Thread(
            target=_qubit_decomposition_impl,
            args=(
                circuit,
                neg_gates,
                neg_state,
                optimality_level,
                multi_thread,
                verbose_level,
                cnot_limit,
            ),
        )

        thread_pos.start()
        thread_neg.start()

        thread_pos.join()
        thread_neg.join()

    else:
        _qubit_decomposition_impl(
            circuit,
            pos_gates,
            pos_state,
            optimality_level,
            multi_thread,
            verbose_level,
            cnot_limit,
        )
        _qubit_decomposition_impl(
            circuit,
            neg_gates,
            neg_state,
            optimality_level,
            multi_thread,
            verbose_level,
            cnot_limit,
        )

    for gate in pos_gates:
        controlled_gate = to_controlled_gate(gate, pivot_qubit, True)
        gates.append(controlled_gate)

    for gate in neg_gates:
        controlled_gate = to_controlled_gate(gate, pivot_qubit, False)
        gates.append(controlled_gate)

    for gate in support_gates:
        gates.append(gate)


def qubit_decomposition(
    circuit: QCircuit,
    target_state: QState,
    optimality_level: int,
    verbose_level: int,
    cnot_limit: int = None,
):
    """Decompose a circuit into a sequence of single qubit gates and CNOT gates .

    :param circuit: [description]
    :type circuit: QCircuit
    :param state: [description]
    :type state: QState
    :param optimality_level: [description]
    :type optimality_level: int
    :return: [description]
    :rtype: [type]
    """

    gates = []

    _qubit_decomposition_impl(
        circuit,
        gates,
        target_state,
        optimality_level,
        True,
        verbose_level,
        cnot_limit=cnot_limit,
    )

    return gates


MIN_ROTATION_ANGLE_SEPARATION = 1e0


def _rotation_angles_optimization(
    rotation_angles: List[float], control_indices: List[int]
):
    assert len(rotation_angles) == 1 << len(
        control_indices
    ), f"len(rotation_angles) = {len(rotation_angles)}, len(control_indices) = {len(control_indices)}"

    dont_cares = set()

    # get all the don't cares
    for index in range(len(control_indices)):
        # check if it is a don't care
        index_is_dont_care = True
        for rotation_index, rotation_angle in enumerate(rotation_angles):
            reversed_index = rotation_index ^ (1 << index)
            if not np.isclose(
                rotation_angles[reversed_index],
                rotation_angle,
                atol=MIN_ROTATION_ANGLE_SEPARATION,
            ):
                index_is_dont_care = False
                break
        if index_is_dont_care:
            dont_cares.add(index)

    if len(dont_cares) == 0:
        # no optimization is needed
        return rotation_angles, control_indices

    # now we reindex the rotation_entry angles
    # print(f"rotation_angles = {rotation_angles}")
    # print(f"reduced {len(dont_cares)} don't cares, dont_cares = {dont_cares}")

    # prepare the new rotation_entry control indices
    new_control_indices = []
    old_indices = []
    for old_index, control_index in enumerate(control_indices):
        if old_index not in dont_cares:
            new_control_indices.append(control_index)
            old_indices.append(old_index)

    # prepare the new rotation_entry angles
    new_rotation_angles = []
    for new_index in range(1 << len(new_control_indices)):
        old_index = 0
        for i in range(len(new_control_indices)):
            old_index |= ((new_index >> i) & 1) << old_indices[i]
        new_rotation_angles.append(rotation_angles[old_index])

    return new_rotation_angles, new_control_indices


def qubit_decomposition_opt(
    circuit: QCircuit,
    state: QState,
    supports: set,
):
    """Composes a circuit decomposition for a circuit .

    the main difference is that, in this implementation:
        1. we do not synthesize the truth table

    to further improve, maybe we can check the actual supports

    :param circuit: [description]
    :type circuit: QCircuit
    :param state: [description]
    :type state: QState
    :param supports: [description]
    :type supports: set
    :return: [description]
    :rtype: [type]
    """
    # we randomly select a pivot qubit
    pivot = select_informative_qubit(state, supports)

    # we first prepare the rotation_entry table
    # the pivot is the target qubit,
    # and the other qubits are the control qubits
    control_indices = list(set(supports) - {pivot})

    rotation_table = [[0, 0] for _ in range(1 << len(control_indices))]

    for index, weight in state.index_to_weight.items():
        rotation_index: int = 0
        for i, support in enumerate(control_indices):
            if index & (1 << support) != 0:
                rotation_index |= 1 << i

        if index & (1 << pivot) != 0:
            rotation_table[rotation_index][1] = weight
        else:
            rotation_table[rotation_index][0] = weight

    rotation_angles = [0 for _ in range(1 << len(control_indices))]

    for i, rotation_entry in enumerate(rotation_table):
        if rotation_entry[0] == 0:
            rotation_angles[i] = np.pi
        elif rotation_entry[1] == 0:
            rotation_angles[i] = 0
        else:
            rotation_angles[i] = 2 * np.arctan(rotation_entry[1] / rotation_entry[0])

    rotation_angles, control_indices = _rotation_angles_optimization(
        rotation_angles, control_indices
    )

    if len(rotation_angles) == 1:
        ry_gate = RY(rotation_angles[0], circuit.qubit_at(pivot))
        gates = [ry_gate]

    else:
        gate = MCMY(
            rotation_angles,
            [circuit.qubit_at(support) for support in control_indices],
            circuit.qubit_at(pivot),
        )

        gates = [gate]

    # we update the state
    index_to_weight = {index: 0 for index in state.index_set}
    for index, weight in state.index_to_weight.items():
        if index & (1 << pivot) == 0:
            index_to_weight[index] += weight
        else:
            index_to_weight[index ^ (1 << pivot)] = np.sqrt(
                weight**2 + state.index_to_weight.get(index ^ (1 << pivot), 0) ** 2
            )

    new_state = QState(index_to_weight, state.num_qubits)

    return gates, new_state



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-09-10 19:24:34
Last Modified by: Hanyu Wang
Last Modified time: 2023-09-10 22:25:33
"""

import numpy as np

from xyz.circuit import CX
from xyz.circuit import X, RY
from xyz.circuit.qcircuit import QCircuit
from xyz.qstate import QState
from xyz.operator import CXOperator, XOperator

ENABLE_Y_REDUCTION = True


def support_reduction(circuit: QCircuit, state: QState, enable_cnot: bool = True):
    """Apply the reduction to the circuit .

    :param circuit: [description]
    :type circuit: QCircuit
    :param state: [description]
    :type state: QState
    """

    # for collisions
    signature_to_qubits = {}

    signatures = state.get_qubit_signatures()
    const1 = state.get_const1_signature()

    gates = []

    new_state = QState(state.index_to_weight, state.num_qubits)

    for qubit_index, signature in enumerate(signatures):
        # this is already not a support
        if signature == 0:
            continue

        # this is not a support if we use a X gate
        if signature == const1:
            gate = X(circuit.qubit_at(qubit_index))
            gates.append(gate)
            operation = XOperator(qubit_index)
            new_state = operation(new_state)
            continue

        # this is not a support if we use a CNOT gate
        if enable_cnot and signature in signature_to_qubits:
            control_qubit = circuit.qubit_at(signature_to_qubits[signature])
            target_qubit = circuit.qubit_at(qubit_index)
            control_phase = True

            gate = CX(control_qubit, control_phase, target_qubit)
            gates.append(gate)
            operation = CXOperator(
                qubit_index, signature_to_qubits[signature], control_phase
            )
            new_state = operation(new_state)
            continue

        # this is not a support if we use a CNOT gate
        if enable_cnot and signature ^ const1 in signature_to_qubits:
            control_qubit = circuit.qubit_at(signature_to_qubits[signature ^ const1])
            target_qubit = circuit.qubit_at(qubit_index)
            control_phase = False

            gate = CX(control_qubit, control_phase, target_qubit)
            gates.append(gate)
            operation = CXOperator(
                qubit_index, signature_to_qubits[signature ^ const1], control_phase
            )
            new_state = operation(new_state)
            continue

        signature_to_qubits[signature] = qubit_index

    if ENABLE_Y_REDUCTION:
        signatures = new_state.get_qubit_signatures()

        for qubit_index, signature in enumerate(signatures):
            # let us check the Y gate
            theta: float = None
            is_separable: bool = True
            index_to_weight = {index: 0 for index in new_state.index_set}
            for index, _ in new_state.index_to_weight.items():
                reversed_index = index ^ (1 << qubit_index)
                if reversed_index not in new_state.index_to_weight:
                    is_separable = False
                    break

                index0 = index & ~(1 << qubit_index)
                index1 = index | (1 << qubit_index)

                weight0 = new_state.index_to_weight[index0]
                weight1 = new_state.index_to_weight[index1]

                assert not np.isclose(
                    np.sqrt(weight0**2 + weight1**2), 0
                ), f"weight0 = {weight0}, weight1 = {weight1}, state = {new_state}"

                _theta = 2 * np.arctan(weight1 / weight0)

                if theta is None:
                    theta = _theta
                    index_to_weight[index0] = np.sqrt(weight0**2 + weight1**2)
                elif np.isclose(theta, _theta):
                    index_to_weight[index0] = np.sqrt(weight0**2 + weight1**2)
                    continue
                else:
                    is_separable = False
                    break

            if is_separable and theta is not None:
                # print(f"state = {new_state}")
                # we can use the Y gate
                gate = RY(theta, circuit.qubit_at(qubit_index))
                gates.append(gate)
                new_state = QState(index_to_weight, new_state.num_qubits)

    return new_state, gates[::-1]



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-06-20 18:47:08
Last Modified by: Hanyu Wang
Last Modified time: 2023-06-22 15:08:10
"""


class QBit:
    """Class method for creating a bit class ."""

    def __init__(self, index: int) -> None:
        self.index = index

    def __str__(self) -> str:
        return f"q{self.index}"

    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, QBit):
            return self.index == __value.index
        return False



#!/usr/bin/env python
# -*- encoding=utf8 -*-

"""
Author: Hanyu Wang
Created time: 2023-09-09 14:34:48
Last Modified by: Hanyu Wang
Last Modified time: 2023-09-09 14:41:40
"""


from xyz.circuit import X
from xyz.circuit import QCircuit
from xyz.qstate import QState


def ground_state_calibration(circuit: QCircuit, state: QState):
    """Calculates the ground state calibration .

    :param circuit: [description]
    :type circuit: QCircuit
    :param state: [description]
    :type state: QState
    """

    assert state.get_sparsity() == 1

    # get the basis state
    index = list(state.index_set)[0]

    gates = []

    # get the 1s in the index
    for qubit in range(state.num_qubits):
        if index & (1 << qubit) != 0:
            gates.append(X(circuit.qubit_at(qubit)))

    return gates
